/* -*-c++-*- 
 *
 * Copyright (C) 2004 Mekensleep
 *
 *	Mekensleep
 *	24 rue vieille du temple
 *	75004 Paris
 *       licensing@mekensleep.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *  Cedric Pinson <cpinson@freesheep.org>
 *  Loic Dachary <loic@gnu.org>
 */

#ifndef MAX2EXG_H

#include "Max.h"
#include "istdplug.h"
#include "stdmat.h"
#include "decomp.h"
#include "shape.h"
#include "bmmlib.h"
#include "bitmap.h"
#include "interpik.h"
#include <string>
#include "exg/exg_mesh.h"
#include <Python.h>
#include <exg/exg_vector3.h>
#include <exg/exg_vector4.h>
#include <exg/exg_matrix3.h>
#include <exg/exg_matrix4.h>
#include <exg/exg_python_scene.h>
#include <resource.h>

class Max2Exg : public SceneExport
{

  // stats data
  int mNbTransforms;
  int mNbMeshs;
  int mNbLights;
  int mNbCameras;


  FILE* mStream;
  Interface* mInterface;
  std::string mNameOfFile;
  std::string mSceneName;

  void ExportGeomObject(INode* node,const std::string& name);
  TriObject* GetTriObjectFromNode(INode *node, TimeValue t, bool &deleteIt);
  void ExportMesh(INode* node, TimeValue t,exg::Pointer<exg::Mesh> mesh);
  bool ProcessNode(exg::PyNodeGroup pyParent,INode* node);
  TimeValue GetStaticFrame() { return 0;}
  void CreateMaterial(Mtl* mtl,int mtlID, int subNo,std::vector<exg::Material*>& result,std::vector<exg::File*>& files);
  Point3 GetVertexNormal(Mesh* mesh, int faceNo, RVertex* rv);
  exg::MapObjectPointer* CreateEntryMapElement(StdMat* mat,Texmap* subTex,int chl,std::vector<exg::File*>& common);
  void SetupMaterialColor(exg::Material* m,Color* ambient,Color* diffuse,Color* specular,float* transparency);

  bool mRotateInRighthand;

  bool IsMeshExport();
  bool IsSceneExport();


  std::map<std::string,int> mMeshProcessed;

  /**
   *  Scene related members
   */
  typedef std::vector<PyObject*> PyNodePath;

  // store trace of light to set include or exclude list
  // at the end of process
  typedef std::map<PyObject*,INode*> LightList;
  LightList mLightDataForList;


  exg::PyNodeGroup mSceneRoot;
  exg::PyNodeTransform CreateNodeTransform(INode* node);
  void CreateNodeMesh(INode* node,exg::PyNodeMesh& mesh,exg::PyNodeTransform& pivot);
  exg::PyNodeLight CreateNodeLight(INode* node);
  exg::PyNodeCamera CreateNodeCamera(INode* node);
  exg::PyNodeTransform CreateNodeHelper(INode* node);
  exg::PyNodeGroup CreateNodeGroup(INode* node);
  exg::PyNodeGroup CreateRootNode();
  void GetMatrixTransform(INode* node,exg::Matrix4f& _transform,exg::Matrix4f& _pivot,bool& _tid,bool& _pid);

  typedef std::map<void*,void*> NodeFactory;
  NodeFactory mNodeFactory;
  void* FindOrCreateNode(void* node,void* pynode,bool& exist);
  exg::PyNode FindPyMesh(exg::PyNodeGroup& n,const std::string& name);
  void SetLightList();

  bool mIncludeMesh;


  /**
   *  Property node that extract additional informations
   */
  struct PropertyNode {

    enum GlobalExport {
      ALL=0,
      TM_ONLY=1,
      SEPARATE_ITEM=2,
      TM_SEPARATE_ITEM=3,
      NONE,
    };

    void Apply(exg::PyNodeMesh& n);
    void Apply(exg::PyNodeTransform& n);

    bool ExportTransform();
    bool ExportItemSeparately();
    bool ExportItem();

    bool light_lighted;
    bool light_castShadow;
    int global_export;
    int billboard_type;
  };

  void CreateFilenameOfMesh(const std::string& _name,std::string& res) { CreateNameOfMesh(_name,res); res.append(".emsh");}
  void CreateNameOfMesh(const std::string& _name,std::string& res) { res="mesh_"+_name;}
  void CreateNameOfTransform(const std::string& _name,std::string& res) { res="transform_"+_name;}

  void GetNodeProperty(INode* node,PropertyNode& prop);
  void CreateDummySceneWithMesh(INode* node,const std::string& _nameOfMesh);


  exg::Matrix4f ConvertTransform(const exg::Matrix4f& _mat);


// constructors/destructor
public:
  Max2Exg();
  virtual ~Max2Exg();

//  static int ExportMeshFromMaxscriptCall(const TCHAR *name, const MeshMaxscriptExportParams& _param);


  // interface functions	
public:
  const TCHAR *AuthorName();
  const TCHAR *CopyrightMessage();
  int DoExport(const TCHAR *name, ExpInterface *ei, Interface *i, BOOL suppressPrompts, DWORD options);
  const TCHAR *Ext(int n);
  int ExtCount();
  const TCHAR *LongDesc();
  const TCHAR *OtherMessage1();
  const TCHAR *OtherMessage2();
  const TCHAR *ShortDesc();
  void ShowAbout(HWND hWnd);
  unsigned int Version();
  BOOL SupportsOptions(int ext, DWORD options);


  bool GetIncludeMesh() { return mIncludeMesh;}
  void SetIncludeMesh(bool b) { mIncludeMesh=b;}

};


#endif
