/* -*-c++-*- 
 *
 * Copyright (C) 2004 Mekensleep
 *
 *	Mekensleep
 *	24 rue vieille du temple
 *	75004 Paris
 *       licensing@mekensleep.com
 /*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *  Cedric Pinson <cpinson@mekensleep.org>
 *
 */

#include "max2exg.h"
#include <exg/exg_utils.h>
#include <exg/exg_vector4.h>
#include <exg/exg_visitor_utils.h>
#include <fstream>


void Max2Exg::SetupMaterialColor(exg::Material* m,Color* ambient,Color* diffuse,Color* specular,float* transparency)
{
  if (ambient) {
    Color cm=(*ambient);
    exg::Vector4f c0(cm[0],cm[1],cm[2],1);
    (*m)["ambient"]=new exg::VectorFloat(c0);
#ifdef DEBUG_MESH
    fprintf(mStream,"ambient %f %f %f %f\n",c0[0],c0[1],c0[2],c0[3]);
#endif
  }
  
  if ( diffuse) {
    Color cm=(*diffuse);
    exg::Vector4f c0(cm[0],cm[1],cm[2],1);
    (*m)["diffuse"]=new exg::VectorFloat(c0);
#ifdef DEBUG_MESH
    fprintf(mStream,"diffuse %f %f %f %f\n",c0[0],c0[1],c0[2],c0[3]);
#endif
  }
  
  if (specular) {
    Color cm=(*specular);
    exg::Vector4f c0(cm[0],cm[1],cm[2],1);
    (*m)["specular"]=new exg::VectorFloat(c0);
#ifdef DEBUG_MESH
    fprintf(mStream,"specular %f %f %f %f\n",c0[0],c0[1],c0[2],c0[3]);
#endif
  }
  
  if (transparency && *transparency>0.0) {
    float cm=*transparency;
    exg::VectorFloat* vf=new exg::VectorFloat();
    (*vf).push_back(cm);
    (*m)["transparency"]=vf;
#ifdef DEBUG_MESH
    fprintf(mStream,"transparency %f\n",cm);
#endif
  }
}



exg::MapObjectPointer* Max2Exg::CreateEntryMapElement(StdMat* mat,Texmap* subTex,int chl,std::vector<exg::File*>& common)
{
  if (subTex->ClassID() != Class_ID(BMTEX_CLASS_ID, 0x00))
    return 0;
  
  std::string channel="unknown";
  switch (chl) {
  case ID_AM: channel="ambient"; break;
  case ID_DI: channel="diffuse"; break;
  case ID_SP: channel="specular"; break;
  case ID_SH: channel="shiness"; break;
  case ID_SS: channel="shinessstrengh"; break;
    //		case ID_SI: channel="selfillu";continue; break;
  case ID_OP: channel="opacity"; break;
  case ID_FI: channel="filtercolor"; break;
  case ID_BU: channel="bump"; break;
  case ID_RL: channel="reflect"; break;
  case ID_RR: channel="refract"; break;
  }
  
  
  BitmapInfo myBinfo;
  myBinfo.SetName(((BitmapTex *)subTex)->GetMapName());
  BMMGetFullFilename(&myBinfo); // get true pathname
  std::string strOriginal( myBinfo.Name());

  // create a relative pathname of bitmap
  std::string::size_type beg=strOriginal.rfind("\\");
  if (beg!=std::string::npos)
    beg+=1;
  else
    beg=0;
  std::string str=mSceneName + "/"+ strOriginal.substr(beg);

  exg::File* fileFound=0;
  for (int i=0;i<common.size();i++)
    if (common[i]->GetName() == str) {
      fileFound=common[i];
      break;
    }
  
  if (!fileFound) {
    fileFound=new exg::File;
    fileFound->SetName(str);
    common.push_back(fileFound);

    if (CopyFile(strOriginal.c_str(),str.c_str(),FALSE))
      fprintf(mStream,"Copy %s to %s\n",strOriginal.c_str(),str.c_str());
    else
      fprintf(mStream,"Copy %s to %s failed\n",strOriginal.c_str(),str.c_str());
  }

  int channelIndex=0;
  StdUVGen* uvGen = ((BitmapTex *)subTex)->GetUVGen(); // check if it's envmap
  if (uvGen)
    channelIndex=uvGen->GetMapChannel();
  
  // blend mode to use
  int trans=mat->GetTransparencyType();
  std::string transparencyMode="";
  if (trans==TRANSP_ADDITIVE && channel=="diffuse") // we assign it only for diffuse
    transparencyMode="additive";

  exg::MapObjectPointer* map=exg::CreateMaterialMapElement(channel,fileFound,2,transparencyMode);
  // store channel index from max. It's a tmp data used to track wich channel of uvs in vertexes used.
  // this data will not be saved.
  (*map)["element"]->AsVectorFloat()->push_back(channelIndex); // store channel index from max

  // debug
  return map;
}


void Max2Exg::CreateMaterial(Mtl* mtl,int mtlID, int subNo,std::vector<exg::Material*>& result,std::vector<exg::File*>& files)
{
  int i;
  TimeValue t = GetStaticFrame();
  
  if (!mtl) 
    return;
  
  TSTR className;
  mtl->GetClassName(className);
  

  // We know the Standard material, so we can get some extra info
  if (mtl->ClassID() == Class_ID(DMTL_CLASS_ID, 0)) {
    StdMat* std = (StdMat*)mtl;
    
    exg::Material* material=new exg::Material;
    material->SetName(std::string(mtl->GetName()));
    result.push_back(material);
#ifdef DEBUG_MESH
    fprintf(mStream,"material %s\n",material->GetName().c_str());
#endif

    Color c[3];
    c[0]=std->GetAmbient(GetStaticFrame());
    c[1]=std->GetDiffuse(GetStaticFrame());
    c[2]=std->GetSpecular(GetStaticFrame());
    float t=std->GetXParency(GetStaticFrame());
    SetupMaterialColor(material,&c[0],&c[1],&c[2],&t);
    
    if (std->GetSelfIllum(t)>50)
      (*material)["selfillum"]=0;
    else 
      (*material)["normal"]=0; //means we use normal per vertex
    
    
    if (mtl->NumSubTexmaps())
      (*material)["maps"]=exg::CreateMaterialMaps();
    
    for (i=0; i<mtl->NumSubTexmaps(); i++) {
      Texmap* subTex = mtl->GetSubTexmap(i);
      float amt = 1.0f;
      if (subTex) {
	// If it is a standard material we can see if the map is enabled.
	if (mtl->ClassID() == Class_ID(DMTL_CLASS_ID, 0)) {
	  if (!((StdMat*)mtl)->MapEnabled(i))
	    continue;
	  
	  amt = ((StdMat*)mtl)->GetTexmapAmt(i, 0);
	  if (subTex->ClassID() == Class_ID(BMTEX_CLASS_ID, 0x00)) {
	    exg::MapObjectPointer* map=CreateEntryMapElement(std,subTex,i,files);
	    (*material)["maps"]->AsVectorObjectPointer()->push_back(map);
#ifdef DEBUG_MESH
            fprintf(mStream,"map channel %s file %s nbElements %d\n",
		    (*map)["name"]->AsFile()->GetName().c_str(),
		    (*map)["file"]->AsFile()->GetName().c_str(),
		    (int)*((*map)["element"]->AsVectorFloat()->begin()));
#endif
      	    
	  } else if (subTex->ClassID() == Class_ID(RGBMULT_CLASS_ID, 0x00)) {
	    if ( subTex->NumSubTexmaps() >= 2) {
	      if ( subTex->GetSubTexmap(0)->ClassID() == Class_ID(BMTEX_CLASS_ID, 0x00) ) {
		exg::MapObjectPointer* map=CreateEntryMapElement(std,subTex->GetSubTexmap(0),i,files);
		(*material)["maps"]->AsVectorObjectPointer()->push_back(map);
#ifdef DEBUG_MESH
    fprintf(mStream,"map channel %s file %s nbElements %d\n",
			(*map)["name"]->AsFile()->GetName().c_str(),
			(*map)["file"]->AsFile()->GetName().c_str(),
			(int)*((*map)["element"]->AsVectorFloat()->begin()));
#endif
	      }

              // light map
	      if ( subTex->GetSubTexmap(1)->ClassID() == Class_ID(BMTEX_CLASS_ID, 0x00) ) {
		exg::MapObjectPointer* map=CreateEntryMapElement(std,subTex->GetSubTexmap(0),-1,files);
		(*map)["name"]->AsFile()->SetName("lightmap2d");
		fprintf(mStream,"material %d lightmap2d used\n",i);
		(*material)["maps"]->AsVectorObjectPointer()->push_back(map);
#ifdef DEBUG_MESH
fprintf(mStream,"map channel %s file %s nbElements %d\n",
			(*map)["name"]->AsFile()->GetName().c_str(),
			(*map)["file"]->AsFile()->GetName().c_str(),
			(int)*((*map)["element"]->AsVectorFloat()->begin()));
#endif
	      }
	    }
	  }
	}
      }
    }
  }
    
  if (mtl->NumSubMtls() > 0)  {
    i=0;
#ifdef DEBUG_MESH
    fprintf(mStream,"nb sub material %d\n", mtl->NumSubMtls());
#endif
    if (mtl->ClassID() == Class_ID(BAKE_SHELL_CLASS_ID, 0))
      i=1; // baked shell material we export only the material in viewport

    for (; i<mtl->NumSubMtls(); i++) {
      Mtl* subMtl = mtl->GetSubMtl(i);
      if (subMtl)
	CreateMaterial(subMtl, 0, i,result,files);
    }
  }

}




void Max2Exg::ExportGeomObject(INode* node,const std::string& name)
{
  // The ObjectState is a 'thing' that flows down the pipeline containing
  // all information about the object. By calling EvalWorldState() we tell
  // max to eveluate the object at end of the pipeline.
  ObjectState objState = node->EvalWorldState(0); 

  // The obj member of ObjectState is the actual object we will export.
  if (objState.obj && objState.obj->SuperClassID()==GEOMOBJECT_CLASS_ID) {

    // Targets are actually geomobjects, but we will export them
    // from the camera and light objects, so we skip them here.
    if (objState.obj->ClassID() == Class_ID(TARGET_CLASS_ID, 0))
      return;
	
    exg::Pointer<exg::Mesh> meshExg=new exg::Mesh;
    // mesh name
    meshExg->SetName(name);
#ifdef DEBUG_MESH
    fprintf(mStream,"%s\n", node->GetName());
#endif
    ExportMesh(node, GetStaticFrame(),meshExg);

    if (mRotateInRighthand) {
      exg::Vector3f axis(1,0,0);
      exg::VisitorTransform v( exg::Matrix4f(-PI/2.0,axis) );
      meshExg->Accept(v);
    }

    fprintf(mStream,"Mesh %s Vertexes %d Triangles %d Materials %d Textures %d\n", node->GetName(),meshExg->GetNbPoints(),meshExg->GetNbPolygons(),meshExg->GetNbMaterials(),meshExg->GetNbFiles());

    std::ofstream fl(name.c_str(),std::ios_base::binary);
    exg::Save(meshExg.Get(),fl);
  }
}


TriObject* Max2Exg::GetTriObjectFromNode(INode *node, TimeValue t, bool &deleteIt)
{
  deleteIt = false;
  TSTR mname;
  Object* nobj=node->GetObjectRef();
  nobj->InitNodeName(mname);
  //  fprintf(mStream, "TRUE Name %s\n",mname);
  
  Object *obj = node->EvalWorldState(t).obj;
  if (obj->CanConvertToType(Class_ID(TRIOBJ_CLASS_ID, 0))) { 
    TriObject *tri = (TriObject *) obj->ConvertToType(t,Class_ID(TRIOBJ_CLASS_ID, 0));
    // Note that the TriObject should only be deleted
    // if the pointer to it is not equal to the object
    // pointer that called ConvertToType()
    if (obj != tri) 
      deleteIt = true;
    return tri;
  } else {
    return 0;
  }
}



void Max2Exg::ExportMesh(INode* node, TimeValue t,exg::Pointer<exg::Mesh> meshExg)
{
  Mtl* nodeMtl = node->GetMtl();
  int vx1, vx2, vx3;
  int i,j;
      
  ObjectState os = node->EvalWorldState(t);
  if (!os.obj || os.obj->SuperClassID()!=GEOMOBJECT_CLASS_ID) {
    return; // Safety net. This shouldn't happen.
  }
      
  bool needDel;
  TriObject* tri = GetTriObjectFromNode(node, t, needDel);
  if (!tri) {
    return;
  }

  TSTR mname;
  mname=node->GetName();

  // table from common material and specific mesh material
  std::map<exg::Material*,exg::Material*> tableMaterials;

  std::vector<exg::Material*> materials;
  std::vector<exg::File*> files;

  if (!nodeMtl) {
    materials.push_back(new exg::Material);
    exg::Vector4f c0(1,0,0,1);
    (*materials[0])["diffuse"]=new exg::VectorFloat(c0);
    materials[0]->SetName("default");
  } else
    CreateMaterial(nodeMtl,-1,-1,materials,files);


  Mesh* mesh = &tri->GetMesh();
  mesh->buildNormals();
      
  //	fprintf(mStream, "%s%s {\n",indent.data(),  ID_MESH);
  //	fprintf(mStream, "%s\t%s %d\n",indent.data(), ID_TIMEVALUE, t);
      
  meshExg->SetNbPoints(mesh->getNumVerts());
  std::vector<exg::Vertex*> vertexesCommon;
  vertexesCommon.resize(mesh->getNumVerts());
  meshExg->SetNbPolygons(mesh->getNumFaces());


  for (i=0;i<mesh->getNumVerts();i++)
    vertexesCommon[i]=new exg::Vertex;
      
#ifdef DEBUG_MESH
  fprintf(mStream, "%d vertexes %d faces\n",meshExg->GetNbPoints(), meshExg->GetNbPolygons());
#endif      
  //points
  for (i=0; i<meshExg->GetNbPoints(); i++) { 
    Point3 v = mesh->verts[i];
    exg::Vector3f ev(v[0],v[1],v[2]);
    vertexesCommon[i]->SetPoint(meshExg->GetPoint(i));
    meshExg->GetPoint(i)->SetPosition(ev);
    exg::Vector3f er=meshExg->GetPoint(i)->GetPosition();
#ifdef DEBUG_MESH
    fprintf(mStream, "point %d %f %f %f\n",
	    i,
	    meshExg->GetPoint(i)->GetPosition()[0],
	    meshExg->GetPoint(i)->GetPosition()[1],
	    meshExg->GetPoint(i)->GetPosition()[2]);
#endif
  }

   
  //faces
  int q;
  for (i=0; i<mesh->getNumFaces(); i++) {
    //meshExg->GetPolygon(i)->SetNbVertexes(3);

    // push vertex
    for (q=0;q<3;q++)
      meshExg->GetPolygon(i)->GetProperty("vertexes")->AsVectorObjectPointer()->push_back(vertexesCommon[mesh->faces[i].v[q]]);

    int idxm=mesh->faces[i].getMatID();
  
    exg::Material* current=materials[0];

    if (materials.size()>1) {
      int nbSubm=materials.size();
      idxm=idxm%nbSubm;
      current=materials[idxm];
    }

	
    meshExg->GetPolygon(i)->SetMaterial(current);
	
    // search if the material (current) exist in mesh, if not found then add it
    if (find((*meshExg)["materials"]->AsVectorObjectPointer()->begin(),
	     (*meshExg)["materials"]->AsVectorObjectPointer()->end(),
	     current) == (*meshExg)["materials"]->AsVectorObjectPointer()->end() )
      (*meshExg)["materials"]->AsVectorObjectPointer()->push_back(current);
	
    for (j=0;j<3;j++) {
      exg::Material* mat=meshExg->GetPolygon(i)->GetMaterial();
      exg::Vertex* vtx=meshExg->GetPolygon(i)->GetVertex(j);
      vtx->SetPoint(meshExg->GetPoint(mesh->faces[i].v[j]));
#ifdef DEBUG_MESH
      fprintf(mStream,"polygon %d - point %d\n",i,mesh->faces[i].v[j]);
#endif

      // check if we maps are sync with mesh
      if (mat->find("maps")!=mat->end() &&  !mat->find("maps")->second->AsVectorObjectPointer()->empty()) {
	exg::VectorObjectPointer* mm=mat->find("maps")->second->AsVectorObjectPointer();
	int vs;
        std::vector<int> toRemove;

        for (vs=0;vs<mm->size();vs++) {
	  std::string nmc=(*(*mm)[vs]->AsMapObjectPointer())["name"]->AsFile()->GetName(); // get name of channel
	  int nbvalue=*(*(*mm)[vs]->AsMapObjectPointer())["element"]->AsVectorFloat()->begin(); // get number of elements per vertex
	  int maxIndex=*((*(*mm)[vs]->AsMapObjectPointer())["element"]->AsVectorFloat()->begin()+1); // get channel index for max
	  
          if (!mesh->mapSupport(maxIndex)) {
	    toRemove.push_back(vs);
#ifdef DEBUG_MESH
	    fprintf(mStream, "Channel vertex %d unsupported remove maps %d\n",vs,maxIndex);
#endif
	  }
	}
	int vci=0;
	for (;vci<toRemove.size();vci++)
	  mm->erase(mm->begin()+vci);
      }


      if (mat->find("maps")!=mat->end() && !mat->find("maps")->second->AsVectorObjectPointer()->empty()) {
	exg::VectorObjectPointer* mm=mat->find("maps")->second->AsVectorObjectPointer();
	int vs;

	// loop on maps element
	for (vs=0;vs<mm->size();vs++) {
	  std::string nmc=(*(*mm)[vs]->AsMapObjectPointer())["name"]->AsFile()->GetName(); // get name of channel
	  int nbvalue=*(*(*mm)[vs]->AsMapObjectPointer())["element"]->AsVectorFloat()->begin(); // get number of elements per vertex
	  int maxIndex=*((*(*mm)[vs]->AsMapObjectPointer())["element"]->AsVectorFloat()->begin()+1); // get channel index for max
	      
	  exg::VectorFloat* vf=new exg::VectorFloat(nbvalue);
	  (*vtx)[nmc]=vf;
	      
#ifdef DEBUG_MESH
	  fprintf(mStream,"polygon %d - channel %s - ",i,nmc.c_str());
#endif	      
	  int kk;
	  if (maxIndex==1) {
	    for (kk=0;kk<nbvalue;kk++)
	      (*vf)[kk]=mesh->tVerts[mesh->tvFace[i].t[j]][kk];
	  } else if (maxIndex>1) {
	    for (kk=0;kk<nbvalue;kk++)
	      (*vf)[kk]=mesh->mapVerts(maxIndex)[mesh->mapFaces(maxIndex)[i].t[j]][kk];
	  }
#ifdef DEBUG_MESH
	  for (kk=0;kk<nbvalue;kk++)
	    fprintf(mStream,"%f ",(*vf)[kk]);
	  fprintf(mStream,"\n");
#endif
	}
      }
	  
      if (mat->AsMapObjectPointer()->find("normal")!=mat->AsMapObjectPointer()->end()) {
	Point3 fn;  // Face normal
	Point3 vn;  // Vertex normal
	int  vert;
	  
	// Face and vertex normals.
	// In MAX a vertex can have more than one normal (but doesn't always have it).
	// This is depending on the face you are accessing the vertex through.
	// To get all information we need to export all three vertex normals
	// for every face.
	Face* f = &mesh->faces[i];
	fn = mesh->getFaceNormal(i);
	vert = f->getVert(j);
	vn = GetVertexNormal(mesh, i, mesh->getRVertPtr(vert));
	exg::VectorFloat* nv=new exg::VectorFloat(exg::Vector3f(vn[0],vn[1],vn[2]));
	(*vtx)["normal"]=nv;

#ifdef DEBUG_MESH
	fprintf(mStream,"polygon %d - normal %f %f %f\n",i, (*nv)[0],(*nv)[1],(*nv)[2]);
	fprintf(mStream,"material id %d\n", mesh->faces[i].getMatID());
#endif
      }
    }
  }

  // put common files in mesh
  if (files.size()) {
    //meshExg->SetNbFiles(files.size());
    for (i=0;i<files.size();i++)
      (*meshExg)["files"]->AsVectorObjectPointer()->push_back(files[i]);
  }

  if (needDel)
    delete tri;
}
  
Point3 Max2Exg::GetVertexNormal(Mesh* mesh, int faceNo, RVertex* rv)
{
  Face* f = &mesh->faces[faceNo];
  DWORD smGroup = f->smGroup;
  int numNormals;
  Point3 vertexNormal;
      
  // Is normal specified
  // SPCIFIED is not currently used, but may be used in future versions.
  if (rv->rFlags & SPECIFIED_NORMAL) {
    vertexNormal = rv->rn.getNormal();
  }
  // If normal is not specified it's only available if the face belongs
  // to a smoothing group
  else if ((numNormals = rv->rFlags & NORCT_MASK) && smGroup) {
    // If there is only one vertex is found in the rn member.
    if (numNormals == 1) {
      vertexNormal = rv->rn.getNormal();
    }
    else {
      // If two or more vertices are there you need to step through them
      // and find the vertex with the same smoothing group as the current face.
      // You will find multiple normals in the ern member.
      for (int i = 0; i < numNormals; i++) {
	if (rv->ern[i].getSmGroup() & smGroup) {
	  vertexNormal = rv->ern[i].getNormal();
	}
      }
    }
  }
  else {
    // Get the normal from the Face if no smoothing groups are there
    vertexNormal = mesh->getFaceNormal(faceNo);
  }
      
  return vertexNormal;
}
