/* -*-c++-*- 
 *
 * Copyright (C) 2004 Mekensleep
 *
 *	Mekensleep
 *	24 rue vieille du temple
 *	75004 Paris
 *       licensing@mekensleep.com
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *  Cedric Pinson <cpinson@freesheep.org>
 *
 */

#include "max2exg.h"
#include <exg/exg_utils.h>
#include <exg/exg_vector4.h>
#include <exg/exg_visitor_utils.h>


exg::Matrix4f Max2Exg::ConvertTransform(const exg::Matrix4f& _mat)
{
  //T = R x _mat x R-1

  exg::Matrix4f r(exg::Matrix4f::Rotation(PI*.5,exg::Vector3f(-1,0,0)));
  exg::Matrix4f r1=r;
  r1.Transpose();
   int i;

#ifdef DEBUG_SCENE
  fprintf(mStream,"TEST matrix\n");
  for (i=0;i<4;i++)
    fprintf(mStream,"%f %f %f %f\n",r[i][0],r[i][1],r[i][2],r[i][3]);



  fprintf(mStream,"prev matrix\n");
  for (i=0;i<4;i++)
    fprintf(mStream,"%f %f %f %f\n",_mat[i][0],_mat[i][1],_mat[i][2],_mat[i][3]);
#endif


  exg::Matrix4f res=r*_mat*r1;
#ifdef DEBUG_SCENE
  fprintf(mStream,"after matrix\n");
  for (i=0;i<4;i++)
    fprintf(mStream,"%f %f %f %f\n",res[i][0],res[i][1],res[i][2],res[i][3]);
#endif

  return res;
}



void Max2Exg::PropertyNode::Apply(exg::PyNodeMesh& n) 
{
#ifdef MEKENSLEEP
  if (light_lighted)
    n.SetProperty("light_lighted",(int)(light_lighted==1));
  if (light_castShadow)
    n.SetProperty("light_castShadow",(int)(light_castShadow==1));
#endif
}

void Max2Exg::PropertyNode::Apply(exg::PyNodeTransform& n) 
{
#ifdef MEKENSLEEP
  // new generate a node transform node if there is a billboard
  if (billboard_type) {
    exg::PyNodeTransform at;
    int nbNodes=n.GetNbNodes();
    if (nbNodes) {
      exg::PyNodeTransform pv(n.GetNode(0).object);
      at.AddNode(pv);
      n.RemoveNode(pv);
    }
    n.AddNode(at);
    at.SetProperty("billboard_type",billboard_type);
    std::string a; //="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"; //stupid bug ...:(
    a.reserve(64);
    n.GetName(a);
    a.insert(0,"auto");
    at.SetName(a);
  }
#endif
}

bool Max2Exg::PropertyNode::ExportTransform() 
{
#ifdef MEKENSLEEP
  if (global_export==ALL || global_export==TM_ONLY || global_export==TM_SEPARATE_ITEM)
    return true;
  return false;
#else
  return true;
#endif
}

bool Max2Exg::PropertyNode::ExportItem() 
{
#ifdef MEKENSLEEP
  if (global_export==ALL || global_export==SEPARATE_ITEM || global_export==TM_SEPARATE_ITEM)
    return true;
  return false;
#else
  return true;
#endif
}

bool Max2Exg::PropertyNode::ExportItemSeparately()
{
#ifdef MEKENSLEEP
  if (global_export==TM_SEPARATE_ITEM || global_export==SEPARATE_ITEM)
    return true;
  return false;
#else
  return false;
#endif
}


void Max2Exg::CreateDummySceneWithMesh(INode* node,const std::string& _nameOfMesh)
{
  exg::PyNodeTransform pivot(0);
  exg::PyNodeMesh mesh(0);
  CreateNodeMesh(node,mesh,pivot);

  std::string ss;
  CreateFilenameOfMesh(_nameOfMesh,ss);
  exg::PyNode o(pivot.object);
  if (!pivot.object)
    o.object=mesh.object;

  ss=mSceneName+"/"+ss;
  o.SetName(ss);
  exg::PySaveScene(o,_nameOfMesh+".escn");

  fprintf(mStream,"mesh exported separately as scene name %s\n", _nameOfMesh.c_str());
}


exg::PyNode Max2Exg::FindPyMesh(exg::PyNodeGroup& node,const std::string& name)
{
  int nb=node.GetNbNodes();
  for (int i=0;i<nb;i++) {
    exg::PyNode n=node.GetNode(i);
    if (n.IsNode("NodeGroup") && !n.IsNode("NodeLight")) {
      exg::PyNodeGroup g(n.object);
      if (g.GetName().find(name)!=std::string::npos)
        return exg::PyNode(g.object);
      exg::PyNode m=FindPyMesh(g,name);
      if (m.object)
        return m;
    } else if (n.IsNode("NodeMesh")) {
      exg::PyNodeMesh mm(n.object);
      if (mm.GetName().find(name)!=std::string::npos)
        return n;
    }
  }

  return exg::PyNode(0);
}


exg::PyNodeTransform Max2Exg::CreateNodeHelper(INode* node)
{ 
 
  // We don't really know what kind of helper this is, but by exporting 
  // the Classname of the helper object, the importer has a chance to 
  // identify it. 
  Object* helperObj = node->EvalWorldState(0).obj; 
  std::string name="item";
  if (!helperObj)
    return 0;

  TSTR className; 
  helperObj->GetClassName(className); 
  std::string myClass=std::string(className);
  std::string nodeName=node->GetName();

  exg::PyNodeTransform ntr;
  exg::Matrix4f ntm,ptm;

  if (1 || node->IsGroupHead() && myClass=="Dummy") { // group
    ntr=CreateNodeTransform(node);
    bool tid,pid;
    GetMatrixTransform(node,ntm,ptm,tid,pid);
    ntr.SetMatrix(ntm*ptm);
    //ntr.SetName(nodeName);
  }
  mNbTransforms++;
  //fprintf(mStream,"node helper %sn", nodeName.c_str()); 


#if 0
  bool exist=false;
  exg::PyNode* o=(exg::PyNode*)FindOrCreateNode(node,new exg::PyNode(),exist);
  if (!exist)
    o->SetName(name);
  ntr->AddNode(o);
#endif

#if 0
  if (helperObj) { 
    TimeValue	t = GetStaticFrame(); 
    Matrix3		oTM = node->GetObjectTM(t); 
    Box3		bbox; 
 
    helperObj->GetDeformBBox(t, bbox, &oTM); 
 
    fprintf(mStream,"%s\t%s %s\n", indent.data(), ID_BOUNDINGBOX_MIN, Format(bbox.pmin)); 
    fprintf(mStream,"%s\t%s %s\n", indent.data(), ID_BOUNDINGBOX_MAX, Format(bbox.pmax)); 
  } 
 
  if (GetIncludeAnim()) { 
    ExportAnimKeys(node, indentLevel); 
  } 
#endif

  return ntr;
} 


exg::PyNodeCamera Max2Exg::CreateNodeCamera(INode* node)
{
  INode* target = node->GetTarget(); 

  bool exist=false;
  exg::PyNodeCamera o;
  exg::PyNodeCamera o2((PyObject*)FindOrCreateNode(node,o.object,exist));
  if (exist)
    return o2;

  o.SetName(node->GetName());
 
  exg::Matrix4f ntm;
  exg::Matrix4f ptm;
  bool pid,tid;
  GetMatrixTransform(node,ntm,ptm,tid,pid);
  exg::Matrix4f ctr=ntm*ptm;

  o.SetMatrix(ctr);

  // If we have a target object, export animation keys for the target too. 
  if (target) { 
    GetMatrixTransform(target,ntm,ptm,tid,pid);
    ntm=ntm*ptm;
    exg::Vector4f pt(ntm[0][3],ntm[1][3],ntm[2][3],1);
    o2.SetTarget(pt);
  }

  CameraState cs; 
  TimeValue t = GetStaticFrame(); 
  Interval valid = FOREVER; 
  
  // Get animation range 
  Interval animRange = mInterface->GetAnimRange(); 
  ObjectState os = node->EvalWorldState(t); 
  CameraObject *cam = (CameraObject *)os.obj; 
  cam->EvalCameraState(t,valid,&cs); 
  

  o.SetFov(cs.fov*180.0/PI);
  o.SetNear(cs.nearRange);
  o.SetFar(cs.farRange);

#ifdef DEBUG_SCENE
  fprintf(mStream,"camera near %f\n",cs.nearRange); 
  fprintf(mStream,"camera far %f\n",cs.farRange); 
  fprintf(mStream,"camera fov %f\n",cs.fov); 
#endif

  mNbCameras++;

  fprintf(mStream,"camera %s\n",node->GetName()); 
  return o;
}


void Max2Exg::SetLightList()
{
  TimeValue t = GetStaticFrame(); 
  LightList::iterator i=mLightDataForList.begin();

  exg::PyNodeLight o(0);
  for (;i!=mLightDataForList.end();i++) {

    ObjectState os = (*i).second->EvalWorldState(t); 
    if (!os.obj) 
      continue; 
	 
    o.object=(*i).first;


    GenLight* light = (GenLight*)os.obj; 
    struct LightState ls; 
    Interval valid = FOREVER; 
    Interval animRange = mInterface->GetAnimRange(); 
 
    light->EvalLightState(t, valid, &ls); 

    ExclList* el = light->GetExclList();  // DS 8/31/00 . switched to NodeIDTab from NameTab 
    if (el->Count()) {
      std::string listType;
      if (el->TestFlag(NT_INCLUDE)) { // include list
        listType="include";
        o.SetIncludeList();
      } else {// exclude
        listType="exclude";
        o.SetExcludeList();
      }

      for (int nameid = 0; nameid < el->Count(); nameid++) { 
        INode *n = (*el)[nameid];	// DS 8/31/00 
        if (n) {
          std::string nodeName=n->GetName();
          exg::PyNode found=FindPyMesh(mSceneRoot,nodeName);
          if (found.object) {
            o.AddNode(found);
#ifdef DEBUG_SCENE
            fprintf(mStream,"item %s - %s",listType.c_str(),found.GetName().c_str());
#endif
          }
        }
      }
    }
  }
}


exg::PyNodeLight Max2Exg::CreateNodeLight(INode* node) 
{ 

  TimeValue t = GetStaticFrame(); 
  ObjectState os = node->EvalWorldState(t); 
  if (!os.obj) 
    return 0; 
	 
  GenLight* light = (GenLight*)os.obj; 
  struct LightState ls; 
  Interval valid = FOREVER; 
  Interval animRange = mInterface->GetAnimRange(); 
 
  light->EvalLightState(t, valid, &ls); 
 
  // This is part os the lightState, but it doesn't 
  // make sense to output as an animated setting so 
  // we dump it outside of ExportLightSettings() 
  
  std::string lightType="unknown";
  bool exist=false;
  exg::PyNodeLight o(0);
  exg::PyNodeDirectionalLight ld;
  exg::PyNodePointLight lp;
  exg::PyNodeSpotLight lt;
  switch(ls.type) { 
  case OMNI_LIGHT:
    lightType="point"; 
    o.object=(PyObject*)FindOrCreateNode(node,lp.object,exist); 
    break; 
  case DIR_LIGHT:
    lightType="direction";
    o.object=(PyObject*)FindOrCreateNode(node,ld.object,exist); 
    break; 
  case FSPOT_LIGHT: 
    lightType="spot"; 
    o.object=(PyObject*)FindOrCreateNode(node,lt.object,exist);
    break; 
  } 
  if (exist)
    return o;

  fprintf(mStream, "Light %s %s\n", lightType.c_str(),node->GetName()); 
  if (!o.object) {// light not supported
    return 0;
  }

  o.SetName(node->GetName());

  exg::Matrix4f ntm,ptm;
  bool pid,tid;
  GetMatrixTransform(node,ntm,ptm,pid,tid);
  exg::Matrix4f ltr=ntm*ptm;

  o.SetMatrix(ltr);
  // If we have a target object, export Node TM for the target too. 
  INode* target = node->GetTarget(); 
  if (target) { 
    GetMatrixTransform(target,ntm,ptm,pid,tid); // not used
  }

  // store trace of ligth
  // we do this because during export I can process mesh after light
  // so i can identify it now.
  mLightDataForList[o.object]=node;

  
  // Export light settings for frame 0 
  o.SetColor(exg::Vector3f(ls.color[0],ls.color[1],ls.color[2]));
#ifdef DEBUG_SCENE
  fprintf(mStream,"light color %f %f %f \n",ls.color[0],ls.color[1],ls.color[2]); 
#endif

  mNbLights++;
  return o;
} 



void Max2Exg::GetNodeProperty(INode* node,PropertyNode& prop)
{
#ifdef MEKENSLEEP
  BOOL b;
  int intValue=0;
  BOOL boolValue=0;

  boolValue=0;
  b=node->GetUserPropBool("light_lighted",boolValue);
  prop.light_lighted=boolValue;

  boolValue=0;
  b=node->GetUserPropBool("light_castShadow",boolValue);
  prop.light_castShadow=boolValue;

  intValue=0;
  b=node->GetUserPropInt("global_export",intValue);
  prop.global_export=intValue;

  intValue=0;
  b=node->GetUserPropInt("billboard_type",intValue);
  prop.billboard_type=intValue;
#else
  prop.light_lighted=0;
  prop.light_castShadow=0;
  prop.global_export=0;
  prop.billboard_type=0;
#endif
}


void Max2Exg::CreateNodeMesh(INode* node,exg::PyNodeMesh& _mesh,exg::PyNodeTransform& _pivot)
{
  bool exist=false;
  Object* ref=node->GetObjectRef()->FindBaseObject();

  // get only the pivot matrix, the Transform matrix is assigned in the NodeTransform
  exg::Matrix4f ntr,ptr;
  bool pid,tid;
  GetMatrixTransform(node,ntr,ptr,pid,tid);

  exg::PyNodeTransform pivot;
  pivot.SetName("pivot");
  pivot.SetMatrix(ptr);


  std::string name;
  std::string fname;
  CreateNameOfMesh(node->GetName(),name);
  CreateFilenameOfMesh(node->GetName(),fname);
  std::string* filename=new std::string(fname);
  std::string* res=0;
  res=(std::string*)FindOrCreateNode(ref,filename,exist);
  if (exist)
    delete filename;
  filename=res;
  exg::PyNodeMesh o;
  o.SetName(name);
  fname=mSceneName+"/"+(*filename);
  o.SetFilename(fname);

  if (!tid) {
    mNbTransforms++;
    pivot.AddNode(o);
    _pivot.object=pivot.object;
  }

  _mesh.object=o.object;

  mNbMeshs++;

  fprintf(mStream,"mesh %s - file %s\n",name.c_str(),filename->c_str()); 
//  return pivot;
}


void Max2Exg::GetMatrixTransform(INode* node,exg::Matrix4f& _transform,exg::Matrix4f& _pivot,bool& _tid,bool& _pid)
{
  _tid=false;
  _pid=false;
  Point3 row[4]; 
  exg::Matrix4f mpvt,ma;
  Matrix3 mpvt2,ma2;
  int i;
  // object local matrix
  {
    Matrix3 m; 
    Matrix3 parentTM, nodeTM, localTM;
    nodeTM = node->GetNodeTM(GetStaticFrame());
    INode * parent = node->GetParentNode();
    parentTM = parent->GetNodeTM(GetStaticFrame());


    localTM = nodeTM*Inverse(parentTM);
    localTM.ValidateFlags();
    m=localTM;
    ma2=m;

    // Dump the whole Matrix 
    row[0] = m.GetRow(0); 
    row[1] = m.GetRow(1); 
    row[2] = m.GetRow(2); 
    row[3] = m.GetRow(3); 
 
    ma=exg::Matrix4f( 
		     exg::Vector4f(row[0][0],row[1][0],row[2][0],row[3][0]), 
		     exg::Vector4f(row[0][1],row[1][1],row[2][1],row[3][1]), 
		     exg::Vector4f(row[0][2],row[1][2],row[2][2],row[3][2]), 
		     exg::Vector4f(0,0,0,1) 
		     ); 

#ifdef DEBUG_SCENE
    fprintf(mStream,"Node Transform %s\n",node->GetName()); 
    for (i=0;i<4;i++) 
      fprintf(mStream,"%f \t%f \t%f \t%f\n",ma[i][0],ma[i][1],ma[i][2],ma[i][3]); 
#endif
    if (0 && m.IsIdentity()) // seems to not work
      _tid=true;
  }

  // pivot/offset matrix
  {
    Matrix3 m; 
    Matrix3 tm(1);
    Point3 pos = node->GetObjOffsetPos();
    tm.PreTranslate(pos);
    Quat quat = node->GetObjOffsetRot();
    PreRotateMatrix(tm, quat);
    ScaleValue scaleValue = node->GetObjOffsetScale();
    ApplyScaling(tm, scaleValue);
    tm.ValidateFlags();
    m=tm;
    mpvt2=m;


    row[0] = m.GetRow(0); 
    row[1] = m.GetRow(1); 
    row[2] = m.GetRow(2); 
    row[3] = m.GetRow(3); 

    mpvt=exg::Matrix4f(
		       exg::Vector4f(row[0][0],row[1][0],row[2][0],row[3][0]), 
		       exg::Vector4f(row[0][1],row[1][1],row[2][1],row[3][1]), 
		       exg::Vector4f(row[0][2],row[1][2],row[2][2],row[3][2]), 
		       exg::Vector4f(0,0,0,1) 
		       ); 

    // get pivot
#ifdef DEBUG_SCENE
    fprintf(mStream,"Node offset Transform %s\n",node->GetName()); 
    for (i=0;i<4;i++) 
      fprintf(mStream,"%f \t%f \t%f \t%f\n",mpvt[i][0],mpvt[i][1],mpvt[i][2],mpvt[i][3]); 
#endif
    if (1 & m.IsIdentity())  // seems to not work
      _pid=true;
  }

  if (mRotateInRighthand) {
    _transform=ConvertTransform(ma);
    _pivot=ConvertTransform(mpvt);
  } else {
    _transform=ma;
    _pivot=mpvt;
  }
}

exg::PyNodeGroup Max2Exg::CreateNodeGroup(INode* node)
{
  bool exist=false;
  exg::PyNodeGroup g;
  exg::PyNodeGroup g2((PyObject*)FindOrCreateNode(node,g.object,exist));
  if (exist)
    return g2;

  g.SetName(node->GetName());
  return g;
}



exg::PyNodeTransform Max2Exg::CreateNodeTransform(INode* node)
{
  exg::Matrix4f ntr,ptr;
  bool pid,tid;
  GetMatrixTransform(node,ntr,ptr,pid,tid);
  exg::PyNodeTransform o;

  int i; 
  std::string a="transform_" + std::string(node->GetName());
#if 0
  fprintf(mStream,"Node Transform concatened %s\n",a.c_str()); 
  for (i=0;i<4;i++) 
    fprintf(mStream,"%f \t%f \t%f \t%f\n",ntr[i][0],ntr[i][1],ntr[i][2],ntr[i][3]); 
#endif

  // assign the matrix to the node
  o.SetMatrix(ntr);
  o.SetName(a);
  mNbTransforms++;
  return o;
}

void* Max2Exg::FindOrCreateNode(void* node,void* ref,bool& exist)
{
  exist=false;
  NodeFactory ::iterator it=mNodeFactory.find(node);
  if (it!=mNodeFactory.end()) {
    exist=true;
    return (*it).second;
  }

  mNodeFactory[node]=ref;
  return ref;
}


exg::PyNodeGroup Max2Exg::CreateRootNode()
{
  Control* ambLightCont;
  Control* bgColCont;

  //  mSceneRoot=new exg::PyNodeGroup();
  mSceneRoot.SetName("root");


  // Output ambient light
  Color ambLight = mInterface->GetAmbient(GetStaticFrame(), FOREVER);

  exg::PyNodeAmbientLight ml;
  //  ml=new exg::PyNodeAmbientLight;
  exg::Vector3f c(ambLight[0],ambLight[1],ambLight[2]);
#ifdef DEBUG_SCENE
  fprintf(mStream, "ambient light %f %f %f\n", ambLight[0],ambLight[1],ambLight[2]);
#endif
  ml.SetName("global_ambient");
  ml.SetColor(c);

  mSceneRoot.AddNode(ml);
  return ml;
}
