/* -*-c++-*- 
 *
 * Copyright (C) 2004 Mekensleep
 *
 *	Mekensleep
 *	24 rue vieille du temple
 *	75004 Paris
 *       licensing@mekensleep.com
/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *  Cedric Pinson <cpinson@freesheep.org>
 *
 */

#include <python_bind.h>


static PyObject* module=0; // change that 


PyObject* PyGetClass(const std::string& classname) 
{ 
  
  if (!module) {
    Py_Initialize();
    module = PyImport_ImportModule("exg.scene");
    if(!module)
    	abort();
  }

  if(PyObject_HasAttrString(module, (char*)classname.c_str())) { 
    return PyObject_GetAttrString(module, (char*)classname.c_str()); 
  } else 
    abort(); 
  return 0; 
} 
 
PyObject* PyNewObject(const std::string& classname, const std::string& instancename) 
{ 
  PyObject* pyclass = PyGetClass(classname); 
  PyObject* instance = PyObject_CallFunction(pyclass, "s", (char*)instancename.c_str()); 
  if(!instance) {
    return 0;
  }
  return instance; 
} 

bool PyIsInstance(PyObject* o,const std::string& s)
{
  PyObject* pc= PyGetClass(s);
  return PyObject_IsInstance(o, pc) ? true : false;
}

void PySaveScene(PyNode* node, const std::string& name)
{
  PyObject* function = PyObject_GetAttrString(module, "save");
  if (PyCallable_Check(function)) {
    PyObject_CallFunction(function, "Os", node->object,name.c_str());
  }
}


PyObject* PyNewMatrix(const exg::Matrix3f& m)
{
  exg::Matrix4f l(m);
  return PyNewMatrix(l);
}

PyObject* PyNewMatrix(const exg::Matrix4f& m)
{
  PyObject* l = PyList_New(4);
  for (int i=0;i<4;i++) {
    PyList_SetItem(l,i,PyNewVector(m[i]));
  }

  PyObject* v=PyNewObject("Matrix");
  PyObject_SetAttrString(v, "data", l);
  return v;
}


PyObject* PyNewVector(const exg::Vector3f& vec)
{
  exg::Vector4f t(vec,1.0);
  return PyNewVector(t);
}

PyObject* PyNewVector(const exg::Vector4f& vec)
{
  PyObject* l = PyList_New(4);
  for (int i=0;i<4;i++)
    PyList_SetItem(l,i,PyFloat_FromDouble(vec[i]*1.0));

  PyObject* v=PyNewObject("Vector");
  PyObject_SetAttrString(v, "data", l);
  return v;
}

exg::Vector4f PyGetVector(PyObject* o)
{
  if (!PyList_Check(o))
    abort();

  exg::Vector4f r;
  int s=PyList_Size(o);
  for (int i=0;i<4;i++) {
    PyObject* o1=PyList_GetItem(o,i);
    if (!PyNumber_Check(o1))
      abort();
    r[i]=PyFloat_AsDouble(PyNumber_Float(o1));
  }
  return r;
}

exg::Matrix4f PyGetMatrix(PyObject* o)
{
  if (!PyList_Check(o))
    abort();

  exg::Matrix4f r;
  int s=PyList_Size(o);
  for (int i=0;i<4;i++) {
    PyObject* o1=PyList_GetItem(o,i);
    r[i]=PyGetVector(o1);
  }
  return r;
}




void PyNodeLight::SetColor(const exg::Vector4f& c)
{
  PyObject* cl=PyNewVector(c);
  if(!PyObject_CallMethod(object, "setColor", "O", cl))
    abort();
}

exg::Vector4f PyNodeLight::GetColor()
{
  PyObject* o=PyObject_CallMethod(object, "getColor","");
  if(!o)
    abort();

  return PyGetVector(o);
}



void PyNodeCamera::SetMatrix(const exg::Matrix4f& m)
{
  PyObject* mtx=PyNewMatrix(m);
  if(!PyObject_CallMethod(object, "setMatrix", "O", mtx))
    abort();
}

exg::Matrix4f PyNodeCamera::GetMatrix()
{
  PyObject* o=PyObject_CallMethod(object, "getMatrix","");
  if(!o)
    abort();
  return PyGetMatrix(o);
}


void PyNodeCamera::SetFov(float f)
{
  if(!PyObject_CallMethod(object, "setFov", "O", PyFloat_FromDouble(f*1.0) ))
    abort();
}
float PyNodeCamera::GetFov()
{
  PyObject* o=PyObject_CallMethod(object, "getFov","");
  if (!o || !PyNumber_Check(o))
    abort();
  
  return PyFloat_AsDouble(PyNumber_Float(o));
}

void PyNodeCamera::SetNear(float f)
{
  if(!PyObject_CallMethod(object, "setNear", "O", PyFloat_FromDouble(f*1.0) ))
    abort();
}
float PyNodeCamera::GetNear()
{
  PyObject* o=PyObject_CallMethod(object, "getNear","");
  if (!o || !PyNumber_Check(o))
    abort();
  
  return PyFloat_AsDouble(PyNumber_Float(o));
}


void PyNodeCamera::SetFar(float f)
{
  if(!PyObject_CallMethod(object, "setFar", "O", PyFloat_FromDouble(f*1.0) ))
    abort();
}
float PyNodeCamera::GetFar()
{
  PyObject* o=PyObject_CallMethod(object, "getFar","");
  if (!o || !PyNumber_Check(o))
    abort();
  
  return PyFloat_AsDouble(PyNumber_Float(o));
}





void PyNode::SetName(const std::string& n)
{
  if(!PyObject_CallMethod(object, "setName", "s", n.c_str()))
    abort();
}

std::string PyNode::GetName()
{
  PyObject* p=PyObject_CallMethod(object, "getName", "");
  if(!p)
    abort();

//  if (!PyString_Check(p))
//    abort();
  return std::string(PyString_AsString(p));
}




void PyNodeMesh::SetFilename(const std::string& n)
{
  if(!PyObject_CallMethod(object, "setFilename", "s", n.c_str()))
    abort();
}

std::string PyNodeMesh::GetFilename()
{
  PyObject* p=PyObject_CallMethod(object, "getFilename", 0);
  if(!p)
    abort();

//  if (!PyString_Check(p))
//    abort();
  return std::string(PyString_AsString(p));
}




void PyNodeGroup::AddNode(PyNode* _o)
{
  mNodes.push_back(_o);
  if(!PyObject_CallMethod(object, "addNode", "O", _o->object))
    abort();
}

PyNode* PyNodeGroup::GetNode(int i)
{
  PyObject* o=PyObject_CallMethod(object, "getNode", "i",i);
  if (!o)
    return 0;

  int s=mNodes.size();
  if (s<=i) {
    mNodes.resize(i+1);

    if (!mNodes[i])
      mNodes[i]=new PyNode(o);
  }

  if (mNodes[i]->object!=o)
    mNodes[i]->object=o;

  return mNodes[i];
  
}


void PyNodeTransform::SetMatrix(const exg::Matrix4f& f)
{
  PyObject* mtx=PyNewMatrix(f);
  if(!PyObject_CallMethod(object, "setMatrix", "O", mtx))
    abort();
}

exg::Matrix4f PyNodeTransform::GetMatrix()
{
  PyObject* o=PyObject_CallMethod(object, "getMatrix","");
  if(!o)
    abort();
  return PyGetMatrix(o);
}
