/* -*-c++-*- 
 *
 * Copyright (C) 2004 Mekensleep
 *
 *	Mekensleep
 *	24 rue vieille du temple
 *	75004 Paris
 *       licensing@mekensleep.com
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *  Cedric Pinson <cpinson@freesheep.org>
 *
 */

#ifndef EXG_PYTHON_SCENE_H
#define EXG_PYTHON_SCENE_H

#include <exg/exg_matrix4.h>
#include <exg/exg_matrix3.h>
#include <exg/exg_vector3.h>
#include <exg/exg_vector4.h>
#include <string>
#include <vector>
#include <Python.h>
#include <exg/exg_export.h>

namespace exg 
{

  class PyNode;

  EXG_EXPORT PyObject*  PyNewMatrix(const Matrix3f& m);
  EXG_EXPORT PyObject*  PyNewMatrix(const Matrix4f& m);
  EXG_EXPORT PyObject*  PyNewVector(const Vector3f& vec);
  EXG_EXPORT PyObject*  PyNewVector(const Vector4f& vec);
  EXG_EXPORT PyObject*  PyGetClass(const std::string& classname);
  EXG_EXPORT PyObject*  PyNewObject(const std::string& classname, const std::string& instancename="o"); 
  EXG_EXPORT bool  PyIsInstance(PyObject* o,const std::string& s);
  EXG_EXPORT void  PySaveScene(PyNode& node, const std::string& name);
  EXG_EXPORT void  PyLoadScene(PyNode& node, const std::string& name);

  EXG_EXPORT Vector4f  PyGetVector(PyObject* o);
  EXG_EXPORT Matrix4f  PyGetMatrix(PyObject* o);


  class EXG_EXPORT PyBase
    {
    public:
      PyBase(){}
      PyBase(const std::string& cname) {
	object = PyNewObject(cname); 
      }

      PyObject* object;
    };



  class EXG_EXPORT PyNode : public PyBase
    {

    public:

      PyNode() : PyBase("Node") {}
      PyNode(const std::string& cname) : PyBase(cname) {}
      PyNode(PyObject* o) { object=o;}

      void SetName(const std::string& n);
      std::string GetName();
      void GetName(std::string& _name);

      bool SetProperty(const std::string& name,int value);
      bool GetPropertyInt(const std::string& name,int& value);

      bool IsNode(const std::string& nm);
    };


  class EXG_EXPORT PyNodeGroup : public PyNode
    {
    public:
      PyNodeGroup(): PyNode("NodeGroup"){ }
      PyNodeGroup(PyObject* o):PyNode(o){}
      PyNodeGroup(const std::string& n): PyNode(n){}

      void AddNode(PyNode _o);
      PyNode GetNode(int i);
      void RemoveNode(PyNode _o);

      int GetNbNodes();

    };


  class EXG_EXPORT PyNodeTransform : public PyNodeGroup
    {
    public:
      PyNodeTransform():PyNodeGroup("NodeTransform"){}
      PyNodeTransform(const std::string& n):PyNodeGroup(n){}
      PyNodeTransform(PyObject* o):PyNodeGroup(o){}

      void SetMatrix(const Matrix4f&);
      Matrix4f GetMatrix();
    };

  class EXG_EXPORT PyNodeMesh : public PyNode
    {

    public:
      PyNodeMesh():PyNode("NodeMesh"){ }
      PyNodeMesh(PyObject* o):PyNode(o){}

      void SetFilename(const std::string& n);
      void GetFilename(std::string& s);
    };

  class EXG_EXPORT PyNodeCamera : public PyNode
    {
    public:
      PyNodeCamera():PyNode("NodeCamera"){ }
      PyNodeCamera(PyObject* o):PyNode(o){}

      void SetMatrix(const Matrix4f& m);
      void SetFov(float fov);
      void SetNear(float n);
      void SetFar(float f);
      void SetTarget(const exg::Vector4f& _target);

      Matrix4f GetMatrix();
      float GetFov();
      float GetNear();
      float GetFar();
      exg::Vector4f GetTarget();
    };


  class EXG_EXPORT PyNodeLight : public PyNodeTransform
    {
    public:
      PyNodeLight():PyNodeTransform("NodeLight"){}
      PyNodeLight(const std::string& n):PyNodeTransform(n){}
      PyNodeLight(PyObject* o):PyNodeTransform(o){}

      void SetColor(const Vector4f& c);
      void SetColor(const Vector3f& c) { SetColor(Vector4f(c,1));}
      Vector4f GetColor();

      void SetIncludeList();
      void SetExcludeList();
    };


  class EXG_EXPORT PyNodeDirectionalLight : public PyNodeLight
    {
    public:
      PyNodeDirectionalLight():PyNodeLight("NodeDirectionalLight"){}
      PyNodeDirectionalLight(PyObject* o):PyNodeLight(o){}

      Vector4f GetDirection();
    };

  class EXG_EXPORT PyNodePointLight : public PyNodeLight
    {
    public:
      PyNodePointLight():PyNodeLight("NodePointLight"){}
      PyNodePointLight(PyObject* o):PyNodeLight(o){}

    };

  class EXG_EXPORT PyNodeSpotLight : public PyNodeLight
    {
    public:
      PyNodeSpotLight():PyNodeLight("NodeSpotLight"){}
      PyNodeSpotLight(PyObject* o):PyNodeLight(o){}

      float GetCutoff();

    };

  class EXG_EXPORT PyNodeAmbientLight : public PyNodeLight
    {
    public:
      PyNodeAmbientLight():PyNodeLight("NodeAmbientLight"){}
      PyNodeAmbientLight(PyObject* o):PyNodeLight(o){}

    };


}

#endif
