/* -*-c++-*- 
 *
 * Copyright (C) 2004 Mekensleep
 *
 *	Mekensleep
 *	24 rue vieille du temple
 *	75004 Paris
 *       licensing@mekensleep.com
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *  Cedric Pinson <cpinson@freesheep.org>
 *
 */

#include <exg/exg_python_scene.h>

namespace exg {

  static PyObject* module=0; // change that 


  PyObject* PyGetClass(const std::string& classname) 
  { 
  
    if (!module) {
      Py_Initialize();
      module = PyImport_ImportModule("exg.scene");
      if(!module)
    	abort();
    }

    if(PyObject_HasAttrString(module, (char*)classname.c_str())) { 
      return PyObject_GetAttrString(module, (char*)classname.c_str()); 
    } else 
      abort(); 
    return 0; 
  } 
 
  PyObject* PyNewObject(const std::string& classname, const std::string& instancename) 
  { 
    PyObject* pyclass = PyGetClass(classname); 
    PyObject* instance = PyObject_CallFunction(pyclass, "s", (char*)instancename.c_str()); 
    if(!instance) {
      return 0;
    }
    return instance; 
  } 

  bool PyIsInstance(PyObject* o,const std::string& s)
  {
    PyObject* pc= PyGetClass(s);
    return PyObject_IsInstance(o, pc) ? true : false;
  }

  void PySaveScene(PyNode& node, const std::string& name)
  {
    PyObject* function = PyObject_GetAttrString(module, "save");
    if (PyCallable_Check(function)) {
      PyObject_CallFunction(function, "Os", node.object,name.c_str());
    }
  }


  void PyLoadScene(PyNode& node, const std::string& name)
  {
    PyGetClass("Node"); // load the module
    PyObject* function = PyObject_GetAttrString(module, "load");
    if (PyCallable_Check(function)) {
       node.object=PyObject_CallFunction(function, "s",name.c_str());
    }
  }


  PyObject* PyNewMatrix(const Matrix3f& m)
  {
    Matrix4f l(m);
    return PyNewMatrix(l);
  }

  PyObject* PyNewMatrix(const Matrix4f& m)
  {
    PyObject* l = PyList_New(4);
    for (int i=0;i<4;i++) {
      PyList_SetItem(l,i,PyNewVector(m[i]));
    }

    PyObject* v=PyNewObject("Matrix");
    PyObject_SetAttrString(v, "data", l);
    return v;
  }


  PyObject* PyNewVector(const Vector3f& vec)
  {
    Vector4f t(vec,1.0);
    return PyNewVector(t);
  }

  PyObject* PyNewVector(const Vector4f& vec)
  {
    PyObject* l = PyList_New(4);
    for (int i=0;i<4;i++)
      PyList_SetItem(l,i,PyFloat_FromDouble(vec[i]*1.0));

    PyObject* v=PyNewObject("Vector");
    PyObject_SetAttrString(v, "data", l);
    return v;
  }

  Vector4f PyGetVector(PyObject* o)
  {
    PyObject* o1=PyObject_GetAttrString(o, "data");
    
    if (!PyList_Check(o1))
      abort();

    Vector4f r;
    int s=PyList_Size(o1);
    for (int i=0;i<4;i++) {
      PyObject* o2=PyList_GetItem(o1,i);
      if (!PyNumber_Check(o2))
	abort();
      r[i]=PyFloat_AsDouble(PyNumber_Float(o2));
    }
    return r;
  }

  Matrix4f PyGetMatrix(PyObject* o)
  {
    PyObject* o1=PyObject_GetAttrString(o, "data");
    
    if (!PyList_Check(o1))
      abort();

    Matrix4f r;
    int s=PyList_Size(o1);
    for (int i=0;i<4;i++) {
      PyObject* o2=PyList_GetItem(o1,i);
      r[i]=PyGetVector(o2);
    }
    return r;
  }




  void PyNodeLight::SetColor(const Vector4f& c)
  {
    PyObject* cl=PyNewVector(c);
    if(!PyObject_CallMethod(object, "setColor", "O", cl))
      abort();
  }

  Vector4f PyNodeLight::GetColor()
  {
    PyObject* o=PyObject_CallMethod(object, "getColor","");
    if(!o)
      abort();

    return PyGetVector(o);
  }

  void PyNodeLight::SetIncludeList()
  {
    PyObject* o=PyObject_CallMethod(object, "setInclude","");
  }

  void PyNodeLight::SetExcludeList()
  {
    PyObject* o=PyObject_CallMethod(object, "setExclude","");
  }



  void PyNodeCamera::SetMatrix(const Matrix4f& m)
  {
    PyObject* mtx=PyNewMatrix(m);
    if(!PyObject_CallMethod(object, "setMatrix", "O", mtx))
      abort();
  }

  Matrix4f PyNodeCamera::GetMatrix()
  {
    PyObject* o=PyObject_CallMethod(object, "getMatrix","");
    if(!o)
      abort();
    return PyGetMatrix(o);
  }

  void PyNodeCamera::SetTarget(const Vector4f& m)
  {
    PyObject* mtx=PyNewVector(m);
    if(!PyObject_CallMethod(object, "setTarget", "O", mtx))
      abort();
  }

  Vector4f PyNodeCamera::GetTarget()
  {
    PyObject* o=PyObject_CallMethod(object, "getTarget","");
    if(!o)
      abort();
    return PyGetVector(o);
  }


  void PyNodeCamera::SetFov(float f)
  {
    if(!PyObject_CallMethod(object, "setFov", "O", PyFloat_FromDouble(f*1.0) ))
      abort();
  }
  float PyNodeCamera::GetFov()
  {
    PyObject* o=PyObject_CallMethod(object, "getFov","");
    if (!o || !PyNumber_Check(o))
      abort();
  
    return PyFloat_AsDouble(PyNumber_Float(o));
  }

  void PyNodeCamera::SetNear(float f)
  {
    if(!PyObject_CallMethod(object, "setNear", "O", PyFloat_FromDouble(f*1.0) ))
      abort();
  }
  float PyNodeCamera::GetNear()
  {
    PyObject* o=PyObject_CallMethod(object, "getNear","");
    if (!o || !PyNumber_Check(o))
      abort();
  
    return PyFloat_AsDouble(PyNumber_Float(o));
  }


  void PyNodeCamera::SetFar(float f)
  {
    if(!PyObject_CallMethod(object, "setFar", "O", PyFloat_FromDouble(f*1.0) ))
      abort();
  }
  float PyNodeCamera::GetFar()
  {
    PyObject* o=PyObject_CallMethod(object, "getFar","");
    if (!o || !PyNumber_Check(o))
      abort();
  
    return PyFloat_AsDouble(PyNumber_Float(o));
  }



  bool PyNode::SetProperty(const std::string& name,int value)
  {
    PyObject* o= PyInt_FromLong(value);
    if (PyObject_SetAttrString(object,(char*)name.c_str(),o)==-1)
      return false;
    return true;
  }

  bool PyNode::GetPropertyInt(const std::string& name,int& value)
  {
    if(PyObject_HasAttrString(object,(char*) name.c_str())) {
      value=PyInt_AsLong(PyObject_GetAttrString(object,(char*)name.c_str()));
      return true;
    }
    return false;    
  }

  void PyNode::SetName(const std::string& n)
  {
    if(!PyObject_CallMethod(object, "setName", "s", n.c_str()))
      abort();
  }

  std::string PyNode::GetName()
  {
    PyObject* p=PyObject_CallMethod(object, "getName", "");
    if(!p)
      abort();

    //  if (!PyString_Check(p))
    //    abort();
    return std::string(PyString_AsString(p));
  }

  void PyNode::GetName(std::string& _name)
  {
    PyObject* p=PyObject_CallMethod(object, "getName", "");
    if(!p)
      abort();
    _name=PyString_AsString(p);
  }


  bool PyNode::IsNode(const std::string& nm)
  {
    return PyObject_IsInstance(object, PyGetClass(nm)) ? true : false;
  }


  void PyNodeMesh::SetFilename(const std::string& n)
  {
    if(!PyObject_CallMethod(object, "setFilename", "s", n.c_str()))
      abort();
  }


  void PyNodeMesh::GetFilename(std::string& str)
  {
    PyObject* p=PyObject_CallMethod(object, "getFilename", 0);
    if(!p)
      abort();

    str=PyString_AsString(p);
    //  std::string filename=PyString_AsString(p);
    //  if (!PyString_Check(p))
    //    abort();
  }




  void PyNodeGroup::AddNode(PyNode _o)
  {
//    mNodes.push_back(_o);
    if(!PyObject_CallMethod(object, "addNode", "O", _o.object))
      abort();
  }

  PyNode PyNodeGroup::GetNode(int i)
  {
    PyObject* o=PyObject_CallMethod(object, "getNodes", 0);
    if (!o)
      return 0;

    PyObject* r=PySequence_GetItem(o,i);

#if 0
    int s=mNodes.size();
    if (s<=i) {
      mNodes.resize(i+1);

      if (!mNodes[i])
	mNodes[i]=new PyNode(r);
    }

    if (mNodes[i]->object!=r)
      mNodes[i]->object=r;

    return mNodes[i];
#endif
    return PyNode(r);
  }

  void PyNodeGroup::RemoveNode(PyNode _o)
  {
    if(!PyObject_CallMethod(object, "removeNode", "O", _o.object))
      abort();
  }

  int PyNodeGroup::GetNbNodes()
  {
    PyObject* o=PyObject_CallMethod(object, "getNodes",0);
    if (!o)
      return 0;

    return PySequence_Size(o);
  }


  void PyNodeTransform::SetMatrix(const Matrix4f& f)
  {
    PyObject* mtx=PyNewMatrix(f);
    if(!PyObject_CallMethod(object, "setMatrix", "O", mtx))
      abort();
  }

  Matrix4f PyNodeTransform::GetMatrix()
  {
    PyObject* o=PyObject_CallMethod(object, "getMatrix",0);
    if(!o)
      abort();
    return PyGetMatrix(o);
  }



  Vector4f PyNodeDirectionalLight::GetDirection()
  {
    exg::Matrix4f m=GetMatrix();
    return exg::Vector4f(m[3][0],m[3][1],m[3][2],0);
  }


  float PyNodeSpotLight::GetCutoff()
  {
    PyObject* o=PyObject_CallMethod(object, "getCutoff","");
    return PyFloat_AsDouble(PyNumber_Float(o));
  }



}
